#include <os.h>
#include "output.h"
#include "charmap_8x12.h"

inline void setPixelBuf(void* scrbuf, unsigned x, unsigned y, uint16_t color){
    if(x < SCREEN_WIDTH && y < SCREEN_HEIGHT){
        if(!has_colors){
            uint8_t* p = (uint8_t*) scrbuf + (x >> 1) + (y << 7) + (y << 5);
            if(x%2)
                *p = (*p & 0xF0) | color;
            else
                *p = (*p & 0x0F) | color << 4;
            //*p = (x % 2) ? ((*p & 0b11110000) | color) : ((*p & 0b00001111) | (color << 4));
            //*p = (*p & (0b1111 << (x%2)*4)) | (color << (1 - (x%2))*4);
        }
        else{
            uint16_t* p = (uint16_t*) (scrbuf + (x << 1) + (y << 9) + (y << 7));//(x << 1) + (y << 9) + (y << 7);
            *p = color;
        }
    }
}

void putChar(void* scrbuf, unsigned x, unsigned y, uint8_t chr){
	int i, j, pixelOn;
	for(i = 0; i < CHAR_HEIGHT; i++){
		for(j = 0; j < CHAR_WIDTH; j++){
			pixelOn = charMap_ascii[chr][i] << j ;
			pixelOn = pixelOn & 0x80 ;
			if(pixelOn){
				setPixelBuf(scrbuf, x + j, y + i, 0);
			}
			else{
				setPixelBuf(scrbuf, x + j, y + i, has_colors ? 0xFFFF : 0xF);
			}
		}
	}
}

void putCharColor(void* scrbuf, unsigned x, unsigned y, uint8_t chr, uint16_t bgcolor, uint16_t chrcolor){
	int i, j, pixelOn;
	for(i = 0; i < CHAR_HEIGHT; i++){
		for(j = 0; j < CHAR_WIDTH; j++){
			pixelOn = charMap_ascii[chr][i] << j ;
			pixelOn = pixelOn & 0x80 ;
			if(pixelOn){
				setPixelBuf(scrbuf, x + j, y + i, chrcolor);
			}
			else{
				setPixelBuf(scrbuf, x + j, y + i, bgcolor);
			}
		}
	}
}

void dispString(void* scrbuf, unsigned x, unsigned y, const char* message){//returns how many lines were written
	int l = strlen(message);
	int i;
	unsigned line=0, col=0;
	for (i = 0; i<l; i++){
	    if(message[i] == '\n'){
	        col = 0;
	        ++line;
	    }
	    else if(message[i] == '\t')
	        col += TAB_WIDTH - (col % TAB_WIDTH);
	    else{
            putChar(scrbuf, x+col*CHAR_WIDTH, y+line*CHAR_HEIGHT, message[i]);
            ++col;
	    }
		if (x+col*CHAR_WIDTH >= SCREEN_WIDTH){
			col = 0;
			++line;
		}
		if(y + line*CHAR_HEIGHT >= SCREEN_HEIGHT)//damit nicht speicher hinter screenbuffer berschrieben wird
			break;
	}
}

void dispStringColor(void* scrbuf, unsigned x, unsigned y, const char* message, uint16_t bgcolor, uint16_t chrcolor){//returns how many lines were written
	int l = strlen(message);
	int i;
	unsigned line=0, col=0;
	for (i = 0; i<l; i++){
	    if(message[i] == '\n'){
	        col = 0;
	        ++line;
	    }
	    else if(message[i] == '\t')
	        col += TAB_WIDTH - (col % TAB_WIDTH);
	    else{
            putCharColor(scrbuf, x+col*CHAR_WIDTH, y+line*CHAR_HEIGHT, message[i], bgcolor, chrcolor);
            ++col;
	    }
		if (x+col*CHAR_WIDTH >= SCREEN_WIDTH){
			col = 0;
			++line;
		}
		if(y + line*CHAR_HEIGHT >= SCREEN_HEIGHT)//damit nicht speicher hinter screenbuffer berschrieben wird
			break;
	}
}
