#include <os.h>
#include "input.h"
#include "output.h"

#define CHARDELAY 60//=600ms
#define CURSORDELAY 20

/*roadmap:
    - speicher, ffnen, usw in funktionen auslagern um stabileren und krzeren code zu erhalten
    - menu: ffnen, speichern, speichern unter
    - error checking bei fwrite
    - bei esc nur meldung anzeigen falls nderung seit letztem speichern
    - nspireio anschauen: vl besser geeignet
    - vertical cursor movement
    - disp optimieren: fps sinkt von 100 auf 35 bei vollem bildschirm (timer statt sleep wrde symptome auch beheben)
    - scrollbalken
    - scrolling vl. mit swipen, oder so
    - multilines fixen: scrolling funktioniert nicht korekt: getline & cursorscreenrow ndern
    - beschleunigender input: beim ersten mal ~500ms delay, dann nur noch ~50ms -> cursordelay&chardelay vereinheitlichen
    - warnung falls datei schon existiert -> berschrieben wird
    - sonderzeichenpopup
    - bei esc zustzlich zu speichern & nicht speichern option abbrechen
    - nsdl/nrgblib anschauen vl.
*/

/*
mem:    0   |   1   |   2   |   ...
pos: 0      1       2       3       ...
*/

void dispCursor(uint8_t* scrbuf, int offset, char* displinep){//offset=where is cursor relative to displinep
    int row=0, col=0;
    int i;
    for(i=0; i<offset; i++){
        switch(displinep[i]){
            case '\n':
                row++;
                col=0;
            break;
            case '\t':
                col += TAB_WIDTH - (col % TAB_WIDTH);
            break;
            default: col++;
        }
        if(col >= SCREEN_WIDTH/CHAR_WIDTH){
            row++;
            col = 0;
        }
    }
    for(i=0; i<CHAR_HEIGHT; i++)
        setPixelBuf(scrbuf, col*CHAR_WIDTH, row*CHAR_HEIGHT + i, 0x0);
}

void insert(char* buffer, int position, char ins){
    memmove(buffer + position + 1, buffer + position, strlen(buffer) - position + 1);
    *(buffer + position) = ins;
}

char delchar(char* buffer, int position){
    char deletedchar = buffer[position-1];
    if(position != strlen(buffer))
        memmove(buffer + position - 1, buffer + position, strlen(buffer) - position + 1);
    else
        *(buffer + position - 1) = '\0';
    return deletedchar;
}

char* getline(char* buffer, int position, int cursorscreenrow){//returns pointer to first character of screen containing position in row=cursorscreenrow
    char* restorebuff = malloc(strlen(buffer) + 1);
    memcpy(restorebuff, buffer, strlen(buffer) + 1);
    buffer[position] = '\0';
    char* linepointer = strrchr(buffer, '\n');
    while(cursorscreenrow > 0 && linepointer != NULL){
        *linepointer = '\0';
        linepointer = strrchr(buffer, '\n');
        cursorscreenrow--;
    }
    memcpy(buffer, restorebuff, strlen(restorebuff) + 1);
    free(restorebuff);
    if(linepointer == NULL)
        return buffer;
    linepointer++;
    return linepointer;
}

int linelen(const char* line){//returns how many characters until newline
    int len;
    for(len = 0; line[len] != '\n'; len++);
    return len;
}

int strWidth(const char* str){
    int width = 0;
    int i;
    for(i=0; i<strlen(str); i++){
        if(str[i] == '\t')
            width += TAB_WIDTH;
        else
            width++;
    }
    if(i != 0 && str[i-1] == '\n')//if last char is newline char: don't count it
        width--;
    return width;
}

int main(int argc, char* argv[]){
    uint8_t* scrbuf = (uint8_t*) malloc(SCREEN_BYTES_SIZE);
    memset(scrbuf,0xFF,SCREEN_BYTES_SIZE);

    char* tmpcp;//char pointer used in various places

    FILE* file = NULL;
    char* path;
    if(argc > 0){
        path = argv[0];
        tmpcp = strrchr(path,'/') + 1;
        *tmpcp = '\0';
    }
    else
        path = "\\Documents\\";
    char savepath[256];
    savepath[0] = '\0';

    char* textbuffer = malloc(1024);
    *textbuffer = '\0';
    char inpbuf = '\0';
    char blockkey = '\0';
    int pos = 0;

    int cursorscreenrow = 0;
    char* displinep;

    int chardelaycounter = CHARDELAY;

    uint8_t arrowbuf = 0;
    uint8_t blockarrow = 0;
    int cursordelaycounter = CURSORDELAY;

    //int fps = 0;
    //uint32_t oldtime = 0;

    while(1){
        /*if(oldtime != *(uint32_t*)0x90090000){
            oldtime = *(uint32_t*)0x90090000;
            printf("fps:%i\n",fps);
            fps = 0;
        }
        ++fps;*/

        sleep(10);//->100fps

        if(on_key_pressed() || isKeyPressed(KEY_NSPIRE_ESC)){//change this later: don't quit if menu still open
            wait_no_key_pressed();
            unsigned inp = show_msgbox_2b("", "Save Document?", "Yes", "No");
            if(inp == 1){
                if(strlen(savepath) == 0)
                    file = fopen(strcat(path,"noname.txt.tns"),"a");
                else
                    file = fopen(savepath, "wb");
                fwrite(textbuffer,1,strlen(textbuffer),file);
                fclose(file);
            }
            break;
        }

        if(strlen(textbuffer) % 1024 == 1023){//if buffer full, make it larger
            textbuffer = realloc(textbuffer, strlen(textbuffer) + 1 + 1024);
            if(textbuffer == NULL)
                return 1;
        }

        //open:
        if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_O)){
            wait_no_key_pressed();
            unsigned inp = show_msgbox_2b("Document not saved!", "Save Document?", "Yes", "No");
            if(inp == 1){
                if(strlen(savepath) == 0){
                    wait_no_key_pressed();
                    int inputlen = show_msg_user_input("Save file as", "", path, &tmpcp);
                    if(inputlen == -1 || inputlen > 255)
                        strcpy(savepath, strcat(path,"noname.txt.tns"));
                    else{
                        strcpy(savepath, tmpcp);
                        free(tmpcp);
                    }
                }
                file = fopen(savepath, "wb");
                fwrite(textbuffer,1,strlen(textbuffer),file);
                fclose(file);
            }
            int inputlen = show_msg_user_input("Open File", "", path, &tmpcp);
            if(inputlen == -1)
                break;
            strcpy(savepath, tmpcp);
            free(tmpcp);
            file = fopen(savepath,"rb");
            if(file == NULL)
                break;
            fseek(file, 0, SEEK_END);
            int filelen = ftell(file);
            printf("filelen %i\n",filelen);
            rewind(file);
            free(textbuffer);
            textbuffer = malloc(filelen + 1024 - (filelen%1024));
            textbuffer[0] = '\0';
            fread(textbuffer, 1, filelen, file);
            fclose(file);
            pos = 0;
            cursorscreenrow = 0;
            wait_no_key_pressed();
        }

        //saving:
        if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_S)){
            wait_no_key_pressed();
            if(strlen(savepath) == 0){
                int inputlen = 0;
                inputlen = show_msg_user_input("Save file as", "", path, &tmpcp);
                if(inputlen == -1 || inputlen > 255)
                    show_msgbox("","Document not saved!");
                else{
                    strcpy(savepath,tmpcp);
                    free(tmpcp);
                }
            }
            if(strlen(savepath) != 0){//don't change this to else!!! savepath is altered by first if
                file = fopen(savepath,"wb");
                fwrite(textbuffer,1,strlen(textbuffer),file);
                fclose(file);
            }
            wait_no_key_pressed();
        }

        //save as:
        if(isKeyPressed(KEY_NSPIRE_CTRL) && isKeyPressed(KEY_NSPIRE_SHIFT) && isKeyPressed(KEY_NSPIRE_S)){
            wait_no_key_pressed();
            int inputlen = 0;
            inputlen = show_msg_user_input("Save file", "", path, &tmpcp);
            if(inputlen == -1 || inputlen > 255)
                show_msgbox("", "Document not saved!");
            else{
                strcpy(savepath,tmpcp);
                free(tmpcp);
                file = fopen(savepath,"wb");
                fwrite(textbuffer,1,strlen(textbuffer),file);
                fclose(file);
            }
            wait_no_key_pressed();
        }

        //input(alphanum/newline/tab/del):
        if((inpbuf = readc()) != '\0'){
            if(inpbuf == blockkey && chardelaycounter)//if key hasn't been released and repeated keypress delay isn't over, ignore keypress
                inpbuf = '\0';
            else{
                blockkey = inpbuf;
                chardelaycounter = CHARDELAY;
                if(inpbuf == '\b'){
                    if(strlen(textbuffer) != 0 && pos != 0){
                        if(delchar(textbuffer, pos) == '\n')
                            cursorscreenrow--;
                        pos--;
                    }
                }
                else{
                    if(inpbuf == '\n')
                        cursorscreenrow++;
                    insert(textbuffer, pos, inpbuf);
                    pos++;
                }
            }
        }
        else//if no key pressed <==> if key released
            blockkey = '\0';
        if(chardelaycounter)
            chardelaycounter--;

        //cursor moving:
        if((arrowbuf = readarrow()) != 0){
            if(arrowbuf == blockarrow && cursordelaycounter)
                arrowbuf = 0;
            else{
                blockarrow = arrowbuf;
                cursordelaycounter = CURSORDELAY;
                if(arrowbuf == ARROW_RIGHT_NUM){
                    if(pos < strlen(textbuffer))
                        pos++;
                }
                else if(arrowbuf == ARROW_LEFT_NUM){
                    if(pos > 0)
                        pos--;
                }
            }
        }
        else
            blockarrow = 0;
        if(cursordelaycounter)
            cursordelaycounter--;

        if(cursorscreenrow < 0)
            cursorscreenrow = 0;
        if(cursorscreenrow >= SCREEN_HEIGHT/CHAR_HEIGHT)
            cursorscreenrow = SCREEN_HEIGHT/CHAR_HEIGHT - 1;

        //disp:
        memset(scrbuf, 0xFF, SCREEN_BYTES_SIZE);
        displinep = getline(textbuffer, pos, cursorscreenrow);
        dispString(scrbuf, 0, 0, displinep);
        dispCursor(scrbuf, pos-(displinep-textbuffer), displinep);

        memcpy(SCREEN_BASE_ADDRESS,scrbuf,SCREEN_BYTES_SIZE);
    }
    free(textbuffer);
    free(scrbuf);
    refresh_osscr();//crashes calc
    return 0;
}
